#######################################################################
# Welcome to PeacoQC - CYTO 2022
######################################################################

## This will force the creation of a backup input and output CSV file for additional debugging
debug_input_output <- FALSE

## For debugging to see what FlowJo is passing to the script.
##
## cat("INPUT-IT                = INPUT_IT \n")
## cat("INPUT-MD          		= INPUT_MAD \n")
## cat("FJ-PAR-Cytof            = FJ_PAR_Cytof \n")
## cat("FJ-PAR-QCSAMPLES        = FJ_PAR_QCSAMPLES \n")
## cat("OUT-PATH            	= OUT_PATH \n")
## cat("FJ-DATA-FILE-PATH       = FJ_DATA_FILE_PATH \n")
## cat("FJ-PARAMS-LIST          = FJ_PARAMS_LIST \n")



## INSTALL REQUIRED PACKAGES ##
#######################
# Install function for packages

loadPackages<-function(...){
	libs<-unlist(list(...))
	req<-unlist(lapply(libs,require,character.only=TRUE))
	need<-libs[req==FALSE]
	if(length(need)>0){
		tryCatch({
			install.packages(pkgs=need, repos="https://cloud.r-project.org")
		}, error = function(e){
			tryCatch({
				install.packages(pkgs=need, repos="https://cloud.r-project.org", type = 'binary')
			}, error = function(e){
				dir.create(path = Sys.getenv("R_LIBS_USER"), showWarnings = FALSE, recursive = TRUE)
				install.packages(pkgs=need, lib = Sys.getenv("R_LIBS_USER"), repos="https://cloud.r-project.org", type = 'binary')
			})
		})
		lapply(need,require,character.only=TRUE)
	}
}


loadBioc<-function(...){
	libs<-unlist(list(...))
	req<-unlist(lapply(libs,require,character.only=TRUE))
	need<-libs[req==FALSE]
	if(length(need)>0){
		tryCatch({
			BiocManager::install(need)
		}, error = function(e){
			tryCatch({
				BiocManager::install(need, type = 'binary')
			}, error = function(e){
				BiocManager::install(need, lib = Sys.getenv("R_LIBS_USER"), type = 'binary')
			})
		})
		lapply(need,require,character.only=TRUE)
	}
}

# Try to write a special function for devtools installs:
devPackages<-function(x){
	devtools::install_github(x);
}

# the Real Deal (aka required packages) #
loadPackages(c('BiocManager', 'devtools', 'flowCore'))

if (!require("PeacoQC",character.only=TRUE)){
	tryCatch({
		devPackages('saeyslab/PeacoQC')
		library('PeacoQC')
	}, error = function(e){
		loadBioc('PeacoQC')
	})
}else{
	require('PeacoQC',character.only=TRUE)
}

loadBioc('ComplexHeatmap');


flowFrameFromFlowJoCSV <- function(csvPath, dropCompPrefix=FALSE) {
	data <- read.csv(csvPath, check.names=FALSE)
	cellIdsColumn <- rep(NA, nrow(data))
	cellIds <- data.frame(CellIdDP = cellIdsColumn)
	names(cellIds) <- CellIdDP
	if (CellIdDP %in% names(data)) {
		cellIds <- data[, CellIdDP]
		data <- data[, -which(names(data) %in% c(CellIdDP))]
	}

	colNames <- colnames(data)
	npar <- length(colNames)
	id <- paste("$P",1:npar,sep="")


	maxRange <- 1024
	range <- maxRange + 1
	minRange <- 0

	colNames <- gsub("^\\s+|\\s+$", "", colNames) 	# Trim the leading and tailing whitespaces

	## FlowJo splits parameter names and description by " :: ", e.g., "Comp-PE-Cy5-A :: CD19#PE-Cy5"
	names <- unlist(lapply(1:npar, function(i) {
		unlist(strsplit(colNames[i], " :: "))[1]
	}))
	## FlowJo indicates that a patameter is compensated by the Comp- prefix in its name, we can drop that if undesirable
	if (dropCompPrefix) {
		names <- gsub("^\\Comp-", "", names)
	}
	## Again splitting by " :: " but this time taking the second part (if present) as description
	desc <- unlist(lapply(1:npar, function(i) {
		s <- unlist(strsplit(colNames[i], " :: "))
		if (length(s) >= 2) s[2]
		else NA
	}))

	pars <- new("AnnotatedDataFrame",
	data=data.frame(
	row.names=I(id), name=I(names), desc=I(desc),
	range=range, minRange=minRange, maxRange=maxRange),
	varMetadata=data.frame(
	row.names=I(c("name","desc","range", "minRange", "maxRange")),
	labelDescription=I(c("Name of Parameter", "Description of Parameter", "Range of Parameter", "Minimum Parameter Value after Transforamtion", "Maximum Parameter Value after Transformation"))
	)
	)

	## Build the descripton
	txt <- list()
	txt["FCSversion"] <- "3"
	txt["$PAR"] <- as.character(npar)
	txt["$TOT"] <- as.character(dim(data)[1])
	txt["$BYTEORD"] <- "4,3,2,1"
	txt["$DATATYPE"] <- "F"
	txt["$MODE"] <- "L"
	txt["$NEXTDATA"] <- "0"

	txt["$BEGINANALYSIS"] <- "0"
	txt["$ENDANALYSIS"] <- "0"
	txt["$BEGINSTEXT"] <- "0"
	txt["$ENDSTEXT"] <- "0"
	dataStart <- 1000 ## This doesn't really matter, it's not an FCS file anyway, but let's provide something reasonable
	txt["$BEGINDATA"] <- as.character(dataStart)
	txt["$ENDDATA"] <- as.character(dataStart + npar * dim(data)[1] * 4 - 1) ## $DATATYPE is F and PnB is 32, so here we do: start + rows x cols x 4 bytes per value - 1

	txt[paste(id,"N",sep="")] <- names
	txt[paste(id,"S",sep="")] <- desc
	txt[paste(id,"R",sep="")] <- as.character(range)
	txt[paste(id,"B",sep="")] <- "32"
	txt[paste(id,"E",sep="")] <- "0,0"
	txt[paste(id,"G",sep="")] <- "1"
	txt[paste("flowCore_",id,"Rmin",sep="")] <- as.character(minRange)
	txt[paste("flowCore_",id,"Rmax",sep="")] <- as.character(maxRange)

	txt["$FIL"] <- basename(csvPath)
	txt["FILENAME"] <- csvPath
	txt["$SYS"] <- paste(Sys.info()['sysname'], Sys.info()['release'])
	txt["CREATOR"] <- R.version.string
	txt["transformation"] <- "applied"
	txt["GUID"] <- generateGUID(data)
	txt["ORIGINALGUID"] <- generateGUID(data)

	txt["$INST"] <- "Unknown institution"
	txt["$OP"] <- "Unknown operator"
	txt["$CYT"] <- "Unknown cytometer"
	txt["$SRC"] <- "Unknown source"

	curtime <- Sys.time()
	txt["$DATE"] <- toupper(format(curtime, "%d-%b-%Y")) ## Current data in the format of "05-OCT-2017"
	timeDelta <- 120 ## picking 2 minutes as cytometer run time
	if ("time" %in% tolower(names)) {
		txt["$TIMESTEP"] <- "0.01" ## We don't know that, so just something that would be reasonable-ish
		timeIndex <- which(tolower(names) == "time")
		timeDelta <- ceiling((max(data[,timeIndex]) - min(data[,timeIndex])) * 0.01)
	}
	if (toupper(format(curtime - timeDelta, "%d-%b-%Y")) == toupper(format(curtime, "%d-%b-%Y"))) {
		## Subtracting timeDelta is still the same day, so let's do it that way.
		txt["$BTIM"] <- format(curtime - timeDelta, "%H:%M:%S")
		txt["$ETIM"] <- format(curtime, "%H:%M:%S")
	} else {
		## Subtracting timeDelta gets us to yesterday, that's not good, so let's just say we started at midnight and add delta to it as ETIM...
		## (all this is arbitrary anyway, we just want resonable values)
		txt["$BTIM"] <- "00:00:00"
		txt["$ETIM"] <- sprintf("%02i:%02i:%02i", floor(timeDelta / 3600), floor((timeDelta%%3600) / 60), floor(timeDelta%%60))
	}
	colnames(data) <- names
	list(fcs=new("flowFrame", exprs=as.matrix(data), parameters=pars, description=txt), cellIds=cellIds)
}

## Generates GUID like "47689979-e471-47a6-aa1a-7910d2d42b03"
## As a side effect, this also sets the seed based in the data passed in
generateGUID <- function(data) {
	mySeed <- round(1000*mean(as.matrix(data))) + dim(data)[1] + dim(data)[2]
	while (mySeed > .Machine$integer.max) mySeed <- round(mySeed/3)
	set.seed(as.integer(mySeed))

	part1 <- paste(unlist(lapply(1:8, function(x) { sprintf("%x", round(runif(1, 0, 9))) })), collapse="", sep="")
	part2 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	part3 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	part4 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	part5 <- paste(unlist(lapply(1:12, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
	paste(part1, part2, part3, part4, part5, sep="-")
}


######################
######################
## Last tested with R 4.2.0 (2022-06-05)
## Check what the user has and compare to what we know works
sessionInfo()

#Inputs from Java
inputIT = INPUT_IT
inputMAD = INPUT_MAD
cytof = FJ_PAR_Cytof
qcSamples = FJ_PAR_QCSAMPLES
OutputPath = "OUT_PATH"
RMargins = FJ_PAR_RMARGINS
EventNumberDP <- "EventNumberDP"
CellIdDP <- "CellID"
csvPATH = "FJ_DATA_FILE_PATH";

####### Based on PeacoQC 1.7.0

tmpRet = flowFrameFromFlowJoCSV(csvPATH);
inputFCS <- tmpRet$fcs



eventsCount <- length(inputFCS@exprs[,1])
parNames <- c(FJ_PARAMS_LIST)

## In CSV files, the parameter names are often like FJComp-xxx while in parNames we may be getting Comp-xxx
if (any(grepl("FJComp", parNames)) && !any(grepl("FJComp", colnames(inputFCS)))) {
	## If it looks like there is FJComp-xxx in parNames but no FJComp-xxx in the column names of the FCS file, then
	## rename FJComp-xxx to Comp-xxx in the parNames and we will be looking for those instead.
	parNames <- gsub("^\\FJComp-", "Comp-", parNames)
}

FCSParams <- c(inputFCS@parameters@data[["name"]])

if (length(parNames) == 0)
stop("The input FCS file does not contain the provided input parameters.", call.=FALSE)



# FlowJo sometimes gives us names like [V705-A] but the actual names in the FCS file are <V705-A>
# FlowJo also replaces '_' with '/' so let's try fix that as well
parNames <- unlist(lapply(parNames, function(name) {
	if (name %in% colnames(inputFCS)) {
		which(FCSParams %in% name)
	} else {
		# Let's try [] to <>
		name2 <- gsub("[", "<", name, fixed=TRUE)
		name2 <- gsub("]", ">", name2, fixed=TRUE)
		if (name2 %in% colnames(inputFCS)) {
			which(FCSParams %in% name2)
		} else {
			# Previous fix did not do it, _ => / on the original names
			name2 <- gsub("_", "/", name, fixed=TRUE)
			if (name2 %in% colnames(inputFCS)) {
				which(FCSParams %in% name2)
			} else {
				# That did not work either, let's try both [] => on top of the previous fix (_ => /)
				name3 <- gsub("[", "<", name2, fixed=TRUE)
				name3 <- gsub("]", ">", name3, fixed=TRUE)
				if (name3 %in% colnames(inputFCS)) {
					which(FCSParams %in% name3)
				} else {
					## Maybe we read a wrong dataset?
					cat(paste("The input FCS file does not contain the provided input parameter, missing", name, "\n"))
					"TRYDIFFERENTDATASET"
				}
			}
		}
	}
}))



## DEBUG info

inputFCS

parNames

eventsCount

#dataset

range(inputFCS, "data")

## END DEBUG info

if (eventsCount == 0)
stop("R failed to read the input FCS file.", call.=FALSE)

if (length(parNames) == 0)
stop("The input FCS file does not contain the provided input parameters.", call.=FALSE)

# if (!("time" %in% tolower(colnames(inputFCS))))
# stop("The input FCS file does not contain the TIME parameter.", call.=FALSE)

# The PeacoQC graph is expecting a Diva time parameter. Our FlowJo time parameter uses much smaller values so we nee to rescale it to make the graph look better.
tryCatch({
inputFCS@exprs[,'Time'] = inputFCS@exprs[,'Time']*1000
}, error = function(e){
})


originalDir <- getwd()


setwd(dirname("FJ_CSV_OUPUT_FILE"))
#fcsName <- stringr::str_remove(inputFCS@description[["$FIL"]], "..ExtNode.csv")
wd <- setwd(OutputPath)

ChRemoveFS <- setdiff(colnames(inputFCS), parNames)
ChRemoveFS

# All good events will be set to 2000, all bad events to 6000.
labels <- rep(2000, eventsCount)


#### We should add margin removal. To do this, we will need to modify the FlowFrame so that the min of each parameter is always 0 and the max is always 1023.
if(RMargins == TRUE){
	BadMargins <- RemoveMargins(ff=inputFCS, channels=parNames, output="full")
	labels[BadMargins$indices_margins] <- 6000
}

tryCatch({pdf(NULL)}, error = function(e){})

if (cytof == TRUE){
	newFlowFrame <- tryCatch({
		newFlowFrame <- PeacoQC(
		ff=inputFCS,
		channels=parNames,
		determine_good_cells="all",
		save_fcs = qcSamples,
		plot=TRUE,
		IT_limit = inputIT,
		MAD = inputMAD,
		output_directory = wd,
		remove_zeros=TRUE,
		time_units=50000)
	}, error = function(e){
		newFlowFrame <- PeacoQC(
		ff=inputFCS,
		channels=parNames,
		determine_good_cells="all",
		save_fcs = qcSamples,
		plot=FALSE,
		IT_limit = inputIT,
		MAD = inputMAD,
		output_directory = wd,
		remove_zeros=TRUE)
	})
} else {
	newFlowFrame <- tryCatch({
		newFlowFrame <- PeacoQC(
		ff=inputFCS,
		channels=parNames,
		determine_good_cells="all",
		save_fcs = qcSamples,
		plot=TRUE,
		IT_limit = inputIT,
		MAD = inputMAD,
		output_directory = wd)
	}, error = function(e){
		newFlowFrame <- PeacoQC(
		ff=inputFCS,
		channels=parNames,
		determine_good_cells="all",
		save_fcs = qcSamples,
		plot=FALSE,
		IT_limit = inputIT,
		MAD = inputMAD,
		output_directory = wd)

	})
}





#Set all bad events to 6000

labels[which(newFlowFrame$GoodCells == "FALSE" )] <- 6000

labels <- as.matrix(labels)
colnames(labels) <- c("PeacoQC")
write.csv(labels, file="FJ_CSV_OUPUT_FILE", row.names=FALSE)

if (debug_input_output)
file.copy("FJ_CSV_OUPUT_FILE", paste("FJ_CSV_OUPUT_FILE", "backup.csv", sep="."))



# R seems to be saving the .RData when exiting, so let's clean up to at least make that tiny (i.e., empty environment)
rm(list=ls())
